/*
* Copyright dylhoxic
*
* dylhoxic _at_ tuxfamily _dot_ org
*
* This software is a first person shooter game.
*
* This software is governed by the CeCILL version 2 license under French law and
* abiding by the rules of distribution of free software.  You can  use, 
* modify and/ or redistribute the software under the terms of the CeCILL
* license as circulated by CEA, CNRS and INRIA at the following URL
* "http://www.cecill.info". 
*
* As a counterpart to the access to the source code and rights to copy,
* modify and redistribute granted by the license, users are provided only
* with a limited warranty  and the software's author,  the holder of the
* economic rights,  and the successive licensors  have only  limited
* liability. 
*
* In this respect, the user's attention is drawn to the risks associated
* with loading,  using,  modifying and/or developing or reproducing the
* software by the user in light of its specific status of free software,
* that may mean  that it is complicated to manipulate,  and  that  also
* therefore means  that it is reserved for developers  and  experienced
* professionals having in-depth computer knowledge. Users are therefore
* encouraged to load and test the software's suitability as regards their
* requirements in conditions enabling the security of their systems and/or 
* data to be ensured and,  more generally, to use and operate it in the 
* same conditions as regards security. 
*
* The fact that you are presently reading this means that you have had
* knowledge of the CeCILL version 2 license and that you accept its terms.
*/

/*! \file map.c
 * \brief map functions
 */

#include "map.h"

/*! \brief initialize and allocate a map structure */
map_t *map_init(Uint32 nb_smesh, Uint32 nb_light)
{
	map_t *cur_map = NULL;
	Uint32 i;

	/* alloc map */
	cur_map = (map_t *)malloc(sizeof(map_t));
	cur_map->smesh = NULL;
	
	if(!cur_map)
	{
		SDL_SetError("Could not allocate map");
		return NULL;
	}
	
	/* fill nb_smesh and alloc smesh */
	if(nb_smesh > 0)
	{
		cur_map->nsmesh = nb_smesh;
		cur_map->smesh = (amesh_t **)malloc(sizeof(amesh_t *)*nb_smesh);
		cur_map->meshname = (char **)malloc(sizeof(char *)*nb_smesh);

		if(cur_map->smesh == NULL || cur_map->meshname == NULL)
		{
			map_free(cur_map);
			return NULL;
		}
		else
		{
			/* set each static mesh to NULL */
			for(i=0; i<nb_smesh; i++)
			{
				cur_map->smesh[i] = NULL;
				cur_map->meshname[i] = NULL;
			}
		}

		cur_map->nlight = nb_light;
		cur_map->light = (light_t **)calloc(nb_light, sizeof(light_t *));
		cur_map->lightname = (char **)calloc(nb_light, sizeof(char *));
		
		if(cur_map->light == NULL)
		{
			map_free(cur_map);
			cur_map=NULL;
		}
		/*	
		else
		{
			for(i=0; i<nb_light; i++)
			{
				// set each light to NULL
				cur_map->light[i] = NULL;
			}
		}
		*/
	}
	return cur_map;
}

/*! \brief free a map structure */
void map_free(map_t *map)
{
	Uint32 i;
	if( map != NULL )
	{
		if(map->smesh!=NULL)
		{
			for(i=0; i < map->nsmesh; i++)
				if(map->smesh[i] != NULL)
					amesh_free((map->smesh)[i]);
			free(map->smesh);
		}

		if(map->meshname != NULL)
		{
			for(i=0; i < map->nsmesh; i++)
				if(map->meshname[i] != NULL)
					free((map->meshname)[i]);
			free(map->meshname);
		}

		if(map->light!=NULL)
		{
			for(i=0; i < map->nlight; i++)
			{	
				if(map->light[i] != NULL)
					free(map->light[i]);
				if(map->lightname[i] != NULL)
					free(map->lightname[i]);
			}
			free(map->light);
			free(map->lightname);
		}
		free(map);
	}
}

/*! \brief load a map */
map_t *map_load(char *filename)
{
	FILE *map = NULL;
	map_t *cur_map = NULL;
	amesh_t *cur_mesh = NULL;
	char line[HKW_LINE_LENGTH];
	char **split_data = NULL;
	Uint32 i=0;
	Uint32 j=0, k=0, l=0;
	
	#if HKW_DEBUG_MAP
	fprintf(stdout, "loading map : %s\n", filename);
	#endif

	map = file_open(filename, "r+");

	if(!map)
		return NULL;

	/* read file header */
	fgets(line, HKW_LINE_LENGTH, map);

	/* first line must be : Hekaw map */
	if(strncmp(line+1, HKW_MAP_L1, HKW_MAP_L1L))
	{
		SDL_SetError("map header is not correct\n");
		return NULL;
	}

	/* exporter used, mostly blender */
	fgets(line, HKW_LINE_LENGTH, map);

	/* read number of objects & lights in map */
	while(fgets(line, HKW_LINE_LENGTH, map))
	{
		if(isspace(line[0]))
			continue;
		if(line[0]==HKW_FILE_CHAR)
		{
			if(line[1]==HKW_MAP_NBO)
			{
				fgets(line, HKW_LINE_LENGTH, map);
				i=atoi(line);
				fgets(line, HKW_LINE_LENGTH, map);
				j=atoi(line);
				cur_map = map_init(i, j);
				break;
			}
		}
	}

	/* free map if not init */
	if(cur_map==NULL)
		return NULL;

	/* load static mesh */
	for(i=0; i<cur_map->nsmesh; i++)
	{
		/* cur_mesh = amesh_load_data2(map, SDL_FALSE); */

		cur_mesh = amesh_load_d1(map, SDL_FALSE);
		if(cur_mesh != NULL)
		{
			amesh_load_d2(map, cur_mesh);
			(cur_map->smesh)[i] = cur_mesh;
			cur_mesh = NULL;
		}
		else
		{
			amesh_free(cur_mesh);
			map_free(cur_map);
			return NULL;
		}
	}

	/* load lights */
	for(i=0; i < j; i++)
	{
		(cur_map->light)[i] = light_load(map);
		if( (cur_map->light)[i] == NULL)
		{
			SDL_SetError("Could not load light %d", i);
			map_free(cur_map);
			return NULL;
		}
	}

	rewind(map);
	j=0;
	k=0;
	while(fgets(line, HKW_LINE_LENGTH, map))
	{
		if(isspace(line[0]))
			continue;
		if(line[0]==HKW_FILE_CHAR)
		{
			if(line[1]=='b' && line[2]==' ')
			{
				i = string_split(line, &split_data, ' ');
				if(i>=2)
				{

					/* printf("line : %s--\n", line ); */
					str_eol( split_data[1] );
					l = strlen( split_data[1] );

					/* printf("adding : %s--, len : %d\n", split_data[1], i ); */

					cur_map->meshname[j] = (char *)malloc(sizeof(char)*(l+1));
					strncpy( cur_map->meshname[j], split_data[1], l);
					cur_map->meshname[j][l] = '\0';

					/* printf("stored : %s--\n", cur_map->meshname[j] ); */

					j++;
				}
				array_free(i, (void **)split_data);
			}
			else if(line[1]=='s')
			{
				i = string_split(line, &split_data, ' ');
				if(i>=2)
				{
					/* printf("line : %s--\n", line ); */
					str_eol( split_data[1] );
					l = strlen( split_data[1] );
					/* printf("adding : %s--, len : %d\n", split_data[1], i ); */
					cur_map->lightname[k] = (char *)malloc(sizeof(char)*(l+1));
					strncpy( cur_map->lightname[k], split_data[1], l);
					cur_map->lightname[k][l] = '\0';
					/* printf("stored : %s--\n", cur_map->lightname[k] ); */
					k++;
				}
				array_free(i, (void **)split_data);
			}
		}
	}
	
	fclose(map);
	return cur_map;
}

/*! \brief display a map */
void map_display(map_t *map)
{
	Uint32 nb_smesh = map->nsmesh;
	Uint32 i;
	
	for(i=0; i<nb_smesh;i++)
	{
		glEnableClientState( GL_NORMAL_ARRAY );
		glNormalPointer(GL_FLOAT, 0, (map->smesh)[i]->normals );

		glEnableClientState( GL_TEXTURE_COORD_ARRAY );
		glTexCoordPointer(2, GL_FLOAT, 0, (map->smesh)[i]->texcoord );

		glEnableClientState( GL_VERTEX_ARRAY );
		glVertexPointer(3, GL_FLOAT, 0, (map->smesh)[i]->vertices );

		glBindTexture(GL_TEXTURE_2D, (map->smesh)[i]->texture[0]);

		glDrawElements (GL_TRIANGLES, (map->smesh)[i]->nbf * 3, GL_UNSIGNED_INT, (map->smesh)[i]->faces );
	}

	glDisableClientState( GL_NORMAL_ARRAY );
	glDisableClientState( GL_TEXTURE_COORD_ARRAY );
	glDisableClientState( GL_VERTEX_ARRAY );
}

/*! \brief pseudo light selection routine */
void light_selection(map_t *map)
{
	Uint32 i, j;

	#if HKW_DEBUG_LIGHT
	GLUquadric *quadratic;
	#endif

	j = HKW_MAP_MAXL < map->nlight ? HKW_MAP_MAXL : map->nlight;

	for(i=0; i < j; i++)
	{	
		glLightfv(GL_LIGHT0+i, GL_POSITION, map->light[i]->position );
		glLightfv(GL_LIGHT0+i, GL_SPOT_DIRECTION, map->light[i]->spot);

		glLightfv(GL_LIGHT0+i, GL_AMBIENT, map->light[i]->ambient);
		glLightfv(GL_LIGHT0+i, GL_DIFFUSE, map->light[i]->diffuse);
		glLightfv(GL_LIGHT0+i, GL_SPECULAR, map->light[i]->specular);

		glLightf( GL_LIGHT0+i, GL_CONSTANT_ATTENUATION, map->light[i]->att_const ); 
		glLightf( GL_LIGHT0+i, GL_LINEAR_ATTENUATION , map->light[i]->att_lin );
		glLightf( GL_LIGHT0+i, GL_QUADRATIC_ATTENUATION , map->light[i]->att_quad );

		glLightf(GL_LIGHT0+i, GL_SPOT_CUTOFF, map->light[i]->spot_cutoff);
		glLightf(GL_LIGHT0+i, GL_SPOT_EXPONENT, map->light[i]->spot_exponent);
		
		/* display lights as red sphere with spot direction */
		#if HKW_DEBUG_LIGHT
		glTranslatef(map->light[i]->position[0], map->light[i]->position[1], map->light[i]->position[2] );
		
		glDisable(GL_LIGHTING);
		
		glColor3f(1.0f, 0.0f, 0.0f);
		
		quadratic=gluNewQuadric();
		gluQuadricDrawStyle(quadratic, GL_FILL);
		gluQuadricNormals(quadratic, GLU_SMOOTH);
		gluSphere(quadratic, 0.1, 32, 32);
		
		glBegin(GL_LINES);
		glColor3f(1.0f, 0.0f, 0.0f);
		glVertex3f(0.0f, 0.0f, 0.0f);
		glColor3f(0.0f, 1.0f, 0.0f);
		
		glVertex3fv(map->light[i]->spot);
		
		
		glEnd();
		
		glEnable(GL_LIGHTING);
		glTranslatef(-map->light[i]->position[0], -map->light[i]->position[1], -map->light[i]->position[2] );
		
		gluDeleteQuadric(quadratic);
		#endif
	}
}

