/*
* Copyright dylhoxic
*
* dylhoxic _at_ tuxfamily _dot_ org
*
* This software is a first person shooter game.
*
* This software is governed by the CeCILL version 2 license under French law and
* abiding by the rules of distribution of free software.  You can  use, 
* modify and/ or redistribute the software under the terms of the CeCILL
* license as circulated by CEA, CNRS and INRIA at the following URL
* "http://www.cecill.info". 
*
* As a counterpart to the access to the source code and rights to copy,
* modify and redistribute granted by the license, users are provided only
* with a limited warranty  and the software's author,  the holder of the
* economic rights,  and the successive licensors  have only  limited
* liability. 
*
* In this respect, the user's attention is drawn to the risks associated
* with loading,  using,  modifying and/or developing or reproducing the
* software by the user in light of its specific status of free software,
* that may mean  that it is complicated to manipulate,  and  that  also
* therefore means  that it is reserved for developers  and  experienced
* professionals having in-depth computer knowledge. Users are therefore
* encouraged to load and test the software's suitability as regards their
* requirements in conditions enabling the security of their systems and/or 
* data to be ensured and,  more generally, to use and operate it in the 
* same conditions as regards security. 
*
* The fact that you are presently reading this means that you have had
* knowledge of the CeCILL version 2 license and that you accept its terms.
*/

/*! \file mesh.c
 * \brief mesh functions
 */

#include "mesh.h"

extern sh_glsl_t sh0;
extern sh_glsl_t sh1;

/*! \brief initialize a static or animated mesh */
amesh_t *amesh_init(Uint32 nbf, Uint32 nbv, SDL_bool is_anim)
{
	amesh_t *cur_mesh = NULL;
	
	if( !nbf || !nbv )
		return NULL;
	
	cur_mesh = (amesh_t *)malloc(sizeof(amesh_t));

	if(cur_mesh == NULL)
	{
		SDL_SetError("Could not alloc mesh");
		return NULL;
	}

	/* cur_mesh->faces = (Uint32 *)malloc( sizeof(Uint32) * nbf * 3); */
	cur_mesh->faces = (GLuint *)malloc( sizeof(GLuint) * nbf * 3);
	cur_mesh->vertices = (GLfloat *)malloc( sizeof(GLfloat) * nbv * 3);
	cur_mesh->normals = (GLfloat *)malloc( sizeof(GLfloat) * nbv * 3);
	cur_mesh->texcoord = (GLfloat *)malloc( sizeof(GLfloat) * nbv * 2);
	
	if( (cur_mesh->faces == NULL) || (cur_mesh->vertices == NULL) ||
		(cur_mesh->normals == NULL) || (cur_mesh->texcoord == NULL) )
	{
		SDL_SetError("Could not alloc data for mesh");
		amesh_free(cur_mesh);
		return NULL;
	}

	if(is_anim == SDL_TRUE)
	{
		/* alloc and zero weights (important) */
		cur_mesh->mat_idx = (GLuint *)calloc(nbv*3, sizeof(GLuint));
		cur_mesh->weights = (GLfloat *)calloc(nbv*3, sizeof(GLfloat));

		if( (cur_mesh->mat_idx == NULL) || (cur_mesh->weights == NULL) )
		{
			SDL_SetError("Could not alloc weights for mesh");
			amesh_free(cur_mesh);
			return NULL;
		}
	}
	else
	{
		cur_mesh->mat_idx = NULL;
		cur_mesh->weights = NULL;
	}

	cur_mesh->nbf = nbf;
	cur_mesh->nbv = nbv;
	
	cur_mesh->skel = NULL;
	
	return cur_mesh;
}

/*! \brief free a mesh */
void amesh_free(amesh_t *mesh)
{
	if(mesh != NULL)
	{
		if(mesh->faces != NULL)
			free(mesh->faces);
		if(mesh->vertices != NULL)
			free(mesh->vertices);
		if(mesh->normals != NULL)
			free(mesh->normals);
		if(mesh->texcoord != NULL)
			free(mesh->texcoord);
		
		if(mesh->mat_idx != NULL)
			free(mesh->mat_idx);
		if(mesh->weights != NULL)
			free(mesh->weights);
					
		if(mesh->skel != NULL)
			ask_free(mesh->skel);
		
		free(mesh);
	}
}

/*! \brief load a static/animated mesh */
amesh_t *amesh_load(char *filename, SDL_bool is_anim)
{
	FILE *mesh = NULL;
	amesh_t *cur_mesh = NULL;
	char line[HKW_LINE_LENGTH];

	#if HKW_DEBUG_MESH
	fprintf(stdout, "loading mesh file : %s\n", filename);
	#endif

	mesh = file_open(filename, "r+");

	if(!mesh)
		return NULL;

	/* read file header */
	fgets(line, HKW_LINE_LENGTH, mesh);

	/* first line must be : Hekaw mesh */
	if(is_anim == SDL_TRUE)
	{
		if(strncmp(line+1, HKW_AMESH_L1, HKW_AMESH_L1L))
			return NULL;
	}
	else
	{
		if(strncmp(line+1, HKW_MESH_L1, HKW_MESH_L1L))
			return NULL;
	}

	/* exporter used, mostly blender */
	fgets(line, HKW_LINE_LENGTH, mesh);
	
	cur_mesh = amesh_load_d1(mesh, is_anim);
	if(cur_mesh!=NULL)
		amesh_load_d2(mesh, cur_mesh);
	else
	{
		amesh_free(cur_mesh);
		cur_mesh = NULL;
	}

	fclose(mesh);
	return cur_mesh;
}

/*! \brief used by amesh_load() function 
 *
 * read the header of a mesh file
 */
amesh_t *amesh_load_d1(FILE *mesh, SDL_bool is_anim)
{
	int i=0;
	Uint32 nbf, nbv;
	amesh_t *cur_mesh = NULL;
	char line[HKW_LINE_LENGTH];

	while( fgets(line, HKW_LINE_LENGTH, mesh) ) {
		/* continue on empty line */
		if(isspace(line[0]))
			continue;
		if(line[0]==HKW_FILE_CHAR) {
			i++;
			if(i>5)
				break;
			switch(line[1]) {
				case HKW_MESH_NBFV:
					/* init amesh struct */
					nbf = atoi(fgets(line, HKW_LINE_LENGTH, mesh));
					nbv = atoi(fgets(line, HKW_LINE_LENGTH, mesh));
					cur_mesh = amesh_init(nbf, nbv, is_anim);
					if(cur_mesh==NULL)
						return NULL;
					break;
				case HKW_MESH_TEX:
					/* load texture */
					fgets(line, HKW_LINE_LENGTH, mesh);
					str_eol(line);
					if( texture_load( line, cur_mesh->texture) 
						== SDL_FALSE )
					{
						amesh_free(cur_mesh);
						return NULL;
					}
					break;
				case HKW_MESH_SK:
					/* load skeleton file if any */
					fgets(line, HKW_LINE_LENGTH, mesh);
					str_eol(line);
					cur_mesh->skel=ask_load( line );
					if(!cur_mesh->skel)
					{
						amesh_free(cur_mesh);
						return NULL;
					}
					break;
				case HKW_MESH_ENDI:
					return cur_mesh;
					break;
				default:
					break;
			}
		}
	}

	/* reach the end of the file */
	amesh_free(cur_mesh);
	return NULL;
}

/*! \brief used by amesh_load() function 
 *
 *  read data of a mesh file
 */
SDL_bool amesh_load_d2(FILE *mesh, amesh_t *cur_mesh)
{
	Uint32 data_section;
	/* read faces, read vertices, read normals, read texture coord, read weights */
	Uint32 rf=0, rv=0, rn=0, rc=0, rw=0;
	int i;
	char line[HKW_LINE_LENGTH];
	char **split_data = NULL;

	while(fgets(line, HKW_LINE_LENGTH, mesh)) {
		/* continue on empty line */ 
		if(isspace(line[0])) {
			continue;
		}
		/* enter face, vertice.... data section */
		if(line[0]==HKW_FILE_CHAR) {
			switch(line[1]) {
				case HKW_MESH_FACE:
					data_section = 1;
					break;
				case HKW_MESH_VERTICE:
					data_section = 2;
					break;
				case HKW_MESH_NORMAL:
					data_section = 3;
					break;
				case HKW_MESH_UV:
					data_section = 4;
					break;
				case HKW_AMESH_WEIGHT:
					data_section = 5;
					break;
				default:
				case HKW_MESH_ENDF:
					/* end of the file */
					if( rf != cur_mesh->nbf || rv != cur_mesh->nbv 
						|| rn != cur_mesh->nbv || rc != cur_mesh->nbv)
					{
						SDL_SetError("Error while reading data...\n");
						amesh_free(cur_mesh);
						return SDL_FALSE;
					}
					else
					{
						return SDL_TRUE;
					}
					break;
			}
			continue;
		}

		/* fill amesh structure */
		switch(data_section)
		{
			/* face data */
			case 1:
				if(rf>=cur_mesh->nbf)
					break;
				i = string_split(line, &split_data, ' ');
				if(i>=3)
				{
					(cur_mesh->faces)[(rf*3)] = atoi(split_data[0]);
					(cur_mesh->faces)[(rf*3)+1] = atoi(split_data[1]);
					(cur_mesh->faces)[(rf*3)+2] = atoi(split_data[2]);
				}
				rf++;
				break;
			/* vertice data */
			case 2:
				if(rv>=cur_mesh->nbv)
					break;
				i = string_split(line, &split_data, ' ');
				if(i>=3)
				{
					(cur_mesh->vertices)[(rv*3)] = atof(split_data[0]);
					(cur_mesh->vertices)[(rv*3)+1] = atof(split_data[1]);
					(cur_mesh->vertices)[(rv*3)+2] = atof(split_data[2]);
				}
				rv++;
				break;
			/* normal data */
			case 3:
				if(rn>=cur_mesh->nbv)
					break;
				i = string_split(line, &split_data, ' ');
				if(i>=3)
				{
					(cur_mesh->normals)[(rn*3)] = atof(split_data[0]);
					(cur_mesh->normals)[(rn*3)+1] = atof(split_data[1]);
					(cur_mesh->normals)[(rn*3)+2] = atof(split_data[2]);
				}
				rn++;
				break;
			/* texture coord data */
			case 4:
				if(rc>=cur_mesh->nbv)
					break;
				i = string_split(line, &split_data, ' ');
				if(i>=2)
				{
					(cur_mesh->texcoord)[(rc*2)] = atof(split_data[0]);
					(cur_mesh->texcoord)[(rc*2)+1] = atof(split_data[1]);
				}
				rc++;
				break;
			/* weight data */
			case 5:
				if(rw>=cur_mesh->nbv)
					break;
				i = string_split(line, &split_data, ' ');

				if( i>=6)
				{
					cur_mesh->mat_idx[ (rw*3)+1 ] = atoi( split_data[3] );
					cur_mesh->weights[ (rw*3)+1 ] = atof( split_data[4] );
				}
				if( i>=8)
				{
					cur_mesh->mat_idx[ (rw*3)+2 ] = atoi( split_data[5] );
					cur_mesh->weights[ (rw*3)+2 ] = atof( split_data[6] );
				}
				if( i>=4)
				{
					cur_mesh->mat_idx[ (rw*3) ] = atoi( split_data[1] );
					cur_mesh->weights[ (rw*3) ] = atof( split_data[2] );
				}
				rw++;
				break;
			default:
				break;	
		}
		/* free the array allocated by string split */
		array_free(i, (void **)split_data);
		split_data=NULL;
	}
	/* should never happen */
	amesh_free(cur_mesh);
	return SDL_FALSE;
}

/*! \brief load a mesh and initialize 'ingame' data */
amesh_ig_t *amesh_ig_init( char *fmesh, SDL_bool is_anim, Uint32 act1, Uint32 act2 )
{
	amesh_ig_t *amesh_ig = NULL;
	
	amesh_ig = (amesh_ig_t *)malloc(sizeof(amesh_ig_t));

	if(!amesh_ig)
	{
		return NULL;
	}

	amesh_ig->mesh = amesh_load(fmesh, is_anim);

	if(!amesh_ig->mesh)
	{
		free(amesh_ig);
		return NULL;
	}
	
	amesh_ig->use_static = 1;
	amesh_ig->nact = 0;

	if( is_anim )
	{
		/* max 2 simultaneous action */	
		amesh_ig->act = (int *)calloc( 2, sizeof(int) );
		amesh_ig->cur_frame = (Uint32 *)calloc( 2, sizeof(Uint32) );
		if( !amesh_ig->act && !amesh_ig->cur_frame )
		{
			return NULL;
		}
		
		if( (act1 > 0) && (act1 <= amesh_ig->mesh->skel->nactions) )
		{
			amesh_ig->nact++;
			amesh_ig->act[0] = act1 - 1;
		}

		if( (act2 > 0) && (act2 <= amesh_ig->mesh->skel->nactions) )
		{
			amesh_ig->nact++;
			amesh_ig->act[1] = act2 - 1;
		}
	
		amesh_ig->sh_sk = (sh_glsl_t *)malloc( sizeof(sh_glsl_t) );

		if( shader_create( amesh_ig->sh_sk , 
					"data/shader/glsl_skinning.vsh", "data/shader/glsl_ppl.fsh" ) == GL_FALSE )
		{
			return NULL;
		}

		amesh_ig->cur_sh = amesh_ig->sh_sk;
		if( amesh_shader( amesh_ig  ) == SDL_FALSE )
		{
			return NULL;
		}
	}
	else
	{
		amesh_ig->uniform_loc = NULL;
		amesh_ig->attrib_loc = NULL;
		amesh_ig->sh_sk = NULL;
		amesh_ig->act = NULL;
		amesh_ig->cur_frame = NULL;
	}

	amesh_ig->sh_static = &sh1;
	amesh_ig->cur_sh = &sh1;	

	return amesh_ig;
}

/*! \brief free an amesh 'ingame' structure */
void amesh_ig_free( amesh_ig_t *mesh_ig  )
{
	amesh_free(mesh_ig->mesh);

	if( mesh_ig->cur_frame != NULL )
		free(mesh_ig->cur_frame);

	if( mesh_ig->act != NULL )
		free(mesh_ig->act);
	
	if( mesh_ig->uniform_loc != NULL )
		free(mesh_ig->uniform_loc);

	if( mesh_ig->attrib_loc != NULL )
		free(mesh_ig->attrib_loc);
	
	if( mesh_ig->sh_sk != NULL )
		free(mesh_ig->sh_sk);
	
	free( mesh_ig );
}

/*! \brief display a mesh */
void amesh_display(amesh_ig_t *mesh_ig)
{	
	amesh_t *amesh=mesh_ig->mesh;

	/* use shader */
	_glUseProgramObjectARB(mesh_ig->cur_sh->program);

	/* enable vertex array */	
	glEnableClientState( GL_NORMAL_ARRAY );
	glNormalPointer(GL_FLOAT, 0, amesh->normals );

	glEnableClientState( GL_TEXTURE_COORD_ARRAY );
	glTexCoordPointer(2, GL_FLOAT, 0, amesh->texcoord );

	glEnableClientState( GL_VERTEX_ARRAY );
	glVertexPointer(3, GL_FLOAT, 0, amesh->vertices );

	if(mesh_ig->use_static == 0)
	{	
		_glEnableVertexAttribArrayARB( mesh_ig->attrib_loc[0] );
		_glEnableVertexAttribArrayARB( mesh_ig->attrib_loc[1] );

		_glVertexAttribPointerARB( mesh_ig->attrib_loc[0], 3, GL_UNSIGNED_INT, 0, 0, amesh->mat_idx);	
		_glVertexAttribPointerARB( mesh_ig->attrib_loc[1], 3, GL_FLOAT, 0, 0, amesh->weights);
	}

	/* render */
	glBindTexture(GL_TEXTURE_2D, amesh->texture[0]);
	glDrawElements (GL_TRIANGLES, amesh->nbf*3, GL_UNSIGNED_INT, amesh->faces );	

	/* disable ...*/
	glDisableClientState( GL_NORMAL_ARRAY );
	glDisableClientState( GL_TEXTURE_COORD_ARRAY );

	if( mesh_ig->use_static == 0 )
	{
		_glDisableVertexAttribArrayARB( mesh_ig->attrib_loc[0]);
		_glDisableVertexAttribArrayARB( mesh_ig->attrib_loc[1]);
	}

	glDisableClientState( GL_VERTEX_ARRAY );

	_glUseProgramObjectARB(0);
}

/*! \brief 'link' an amesh and a skinning shader 
 * 
 * get uniform locations : bone matrix 
 * and get attribute locations : weights & matrix index
 */
SDL_bool amesh_shader(amesh_ig_t *mesh_ig)
{
	Uint32 i;
	int action = 0;
	char bonemat[] = HKW_AMESH_UNAME1;

	/* it's better to use the program on some cards...*/
	_glUseProgramObjectARB(mesh_ig->cur_sh->program);

	/* alloc for two attribs adress */
	mesh_ig->attrib_loc = (GLint *)malloc(sizeof(GLint)*2);
	
	if( !mesh_ig->attrib_loc )
	{
		SDL_SetError("Could not alloc attrib variable");
		_glUseProgramObjectARB(0);
		return SDL_FALSE;
	}

	/* get attrib location in shader */
	mesh_ig->attrib_loc[0] = _glGetAttribLocationARB( mesh_ig->cur_sh->program, HKW_AMESH_ANAME1 );
	mesh_ig->attrib_loc[1] = _glGetAttribLocationARB( mesh_ig->cur_sh->program, HKW_AMESH_ANAME2 );

	if( (mesh_ig->attrib_loc[0] == -1) || (mesh_ig->attrib_loc[1] == -1) )
	{	
		SDL_SetError("Could not get attrib location");
		_glUseProgramObjectARB(0);
		return SDL_FALSE;
	}

	/* alloc uniform adress */
	mesh_ig->uniform_loc = (GLint *)malloc(sizeof(GLint)*mesh_ig->mesh->skel->action[action]->nbones);

	if( !mesh_ig->uniform_loc )
	{
		SDL_SetError("Could not alloc uniform pointer");
		_glUseProgramObjectARB(0);
		return SDL_FALSE;
	}
	
	for(i=0;i<mesh_ig->mesh->skel->action[action]->nbones;i++)
	{
		sprintf(bonemat, HKW_AMESH_UNAME1, i);
		mesh_ig->uniform_loc[i] = _glGetUniformLocationARB( mesh_ig->cur_sh->program, bonemat );
		if( mesh_ig->uniform_loc[i] == -1 )
		{
			SDL_SetError("Could not get uniform location");
			_glUseProgramObjectARB(0);
			return SDL_FALSE;
		}		
	}

	/* disable the program */
	_glUseProgramObjectARB(0);
	return SDL_TRUE;
}

/*! \brief update an amesh 
 * 
 * update_framerate() call this function
 */
void amesh_update(amesh_ig_t *amesh, GLfloat u)
{
	switch(amesh->nact)
	{
		case 0 : 
			/* use static shader */
			amesh->cur_sh = amesh->sh_static;
			amesh->use_static = 1;
			break;
		case 1 : 
			amesh->cur_sh = amesh->sh_sk;
			amesh->use_static = 0;
			amesh_update_shader1(amesh, u );
			break;
		case 2 :
			amesh->cur_sh = amesh->sh_sk;
			amesh->use_static = 0;
			amesh_update_shader2(amesh, u );
			break;
	}
}

/*! \brief display a mesh with animation - 1 action */
void amesh_update_shader1(amesh_ig_t *amesh, float u)
{
	GLfloat mat0[16];
	
	Uint32 cur_frame_g = amesh->cur_frame[0];
	Uint32 action_g = amesh->act[0];

	Uint32 i = 0;
	Uint32 nbo_mdl = amesh->mesh->skel->action[action_g]->nbones;
	Uint32 nframe1 = amesh->mesh->skel->action[action_g]->nframes;
	
	quat_t *quat1 = NULL;
	quat_t *quat3 = NULL;
	quat_t temp;
	/*
	quat_t temp2;
	*/

	_glUseProgramObjectARB(amesh->sh_sk->program);

	/* fill the shader with bone matrice (uniform)*/
	for(i=0;i<nbo_mdl;i++)
	{

		quat1 = &((amesh->mesh->skel->action[action_g]->quat)[cur_frame_g][i]);
	
		/* compute quaternion interpolation slerp or nlerp */
		if(u<1.0f)
		{	
			if(cur_frame_g == 0)
			{
				/* u = 0.0f; */
				quat3 = quat1;
			}
			else
			{
				quat3 = &((amesh->mesh->skel->action[action_g]->quat)[(cur_frame_g-1)][i]);
			}

			temp = quat_nlerp(quat1, quat3, u);
			/*
			temp2 = quat_slerp(quat1, quat3, u);
			*/

			quat_tomat( &temp, mat0);
		}
		else
		{
			quat_tomat( quat1, mat0);
		}

		_glUniformMatrix4fvARB( amesh->uniform_loc[ i ], 1, 0, mat0 );
	}

	/* only increase current frame counter when no interpolation */
	if(u == 1.0f)
	{		
		(amesh->cur_frame[0])++;
		if(amesh->cur_frame[0] >= nframe1)
			amesh->cur_frame[0] = 0;	
	}

	_glUseProgramObjectARB(0);
}

/*! \brief display a mesh with animation - 2 simultaneaous action */
void amesh_update_shader2(amesh_ig_t *amesh, float u)
{
	GLfloat mat0[16];
	
	Uint32 cur_frame_g1 = amesh->cur_frame[0];
	Uint32 action_g1 = amesh->act[0];

	Uint32 cur_frame_g2 = amesh->cur_frame[1];
	Uint32 action_g2 = amesh->act[1];

	Uint32 cur_frame_g = cur_frame_g1;
	Uint32 action_g = action_g1;

	Uint32 i = 0;
	Uint32 nbo_mdl = amesh->mesh->skel->action[action_g1]->nbones;
	Uint32 nframe1 = amesh->mesh->skel->action[action_g1]->nframes;
	Uint32 nframe2 = amesh->mesh->skel->action[action_g2]->nframes;

	quat_t *quat1 = NULL;
	
	quat_t *quat3 = NULL;
	quat_t temp;
	/* quat_t temp2; */

	_glUseProgramObjectARB(amesh->sh_sk->program);

	/* fill the shader with bone matrice (uniform) */
	for(i=0;i<nbo_mdl;i++)
	{

		quat1 = &((amesh->mesh->skel->action[action_g1]->quat)[cur_frame_g1][i]);
		
		if( quat_isid( quat1 ) == SDL_TRUE )
		{
			
			cur_frame_g = cur_frame_g2;
			action_g = action_g2;
			
			quat1 = &((amesh->mesh->skel->action[action_g2]->quat)[cur_frame_g2][i]);
		}
		else
		{
			cur_frame_g = cur_frame_g1;
			action_g = action_g1;
		}

		/* compute quaternion interpolation slerp or nlerp */ 
		if(u<1.0f)
		{	
			if(cur_frame_g == 0)
			{
				/* u = 0.0f; */
				quat3 = quat1;
			}
			else
			{
				quat3 = &((amesh->mesh->skel->action[action_g]->quat)[(cur_frame_g-1)][i]);
			}

			temp = quat_nlerp(quat1, quat3, u);
			/* temp2 = quat_slerp(quat1, quat3, u); */

			quat_tomat( &temp, mat0);
		}
		else
		{
			quat_tomat( quat1, mat0);
		}

		_glUniformMatrix4fvARB( amesh->uniform_loc[ i ], 1, 0, mat0 );
	}

	/* only increase current frame counter when no interpolation */ 
	if(u == 1.0f)
	{	
		(amesh->cur_frame[0])++;
		if(amesh->cur_frame[0] >= nframe1)
			amesh->cur_frame[0] = 0;

		(amesh->cur_frame[1])++;
		if(amesh->cur_frame[1] >= nframe2)
			amesh->cur_frame[1] = 0;
	}

	_glUseProgramObjectARB(0);
}

