/*
* Copyright dylhoxic
*
* dylhoxic _at_ tuxfamily _dot_ org
*
* This software is a first person shooter game.
*
* This software is governed by the CeCILL version 2 license under French law and
* abiding by the rules of distribution of free software.  You can  use, 
* modify and/ or redistribute the software under the terms of the CeCILL
* license as circulated by CEA, CNRS and INRIA at the following URL
* "http://www.cecill.info". 
*
* As a counterpart to the access to the source code and rights to copy,
* modify and redistribute granted by the license, users are provided only
* with a limited warranty  and the software's author,  the holder of the
* economic rights,  and the successive licensors  have only  limited
* liability. 
*
* In this respect, the user's attention is drawn to the risks associated
* with loading,  using,  modifying and/or developing or reproducing the
* software by the user in light of its specific status of free software,
* that may mean  that it is complicated to manipulate,  and  that  also
* therefore means  that it is reserved for developers  and  experienced
* professionals having in-depth computer knowledge. Users are therefore
* encouraged to load and test the software's suitability as regards their
* requirements in conditions enabling the security of their systems and/or 
* data to be ensured and,  more generally, to use and operate it in the 
* same conditions as regards security. 
*
* The fact that you are presently reading this means that you have had
* knowledge of the CeCILL version 2 license and that you accept its terms.
*/

/*! \file quaternion.c
 * \brief quaternion functions
 */

#include "quaternion.h"

/*! \brief initialize an identity quaternion
 * \warning does not allocate quaternion
 */
void quat_init_id(quat_t *q1)
{
	q1->x = 0.0f; q1->y = 0.0f; q1->z = 0.0f; q1->w = 1.0f;
}

/*! \brief initialize a quaternion given an angle (degree) && an axis
 * \warning does not allocate quaternion
 */
void quat_init_axisangle( quat_t *quat, float theta, float x, float y, float z )
{
	GLfloat angle, s;

	/* deg --> rad */
	angle = ( theta / 180.0f ) * M_PI;
	s = sin( angle * 0.5f );
	quat->w = cos( angle * 0.5f );
	quat->x = x * s;
	quat->y = y * s;
	quat->z = z * s;
}


/*! \brief return SDL_TRUE if quat is an identity quaternion
 *
 * return SDL_TRUE or SDL_FALSE, use float_equal function
 */
SDL_bool quat_isid( quat_t *quat )
{
	if( ( float_equal(quat->w, 1.0f) == SDL_TRUE )
		|| ( float_equal(quat->w, -1.0f) == SDL_TRUE) )
	{
		if( (float_equal(quat->x, 0.0f) == SDL_TRUE )
			&& (float_equal(quat->y, 0.0f) == SDL_TRUE)
			&& (float_equal(quat->z, 0.0f) == SDL_TRUE) )
		{
			return SDL_TRUE;
		}
	}
	return SDL_FALSE;	
}

/*! \brief return the direction (vector) of a quaternion */
vec_t quat_direction(quat_t *q1)
{
	vec_t direction;
	direction.x = 2.0f * ( q1->x * q1->z - q1->w * q1->y );
	direction.y = 2.0f * ( q1->y * q1->z + q1->w * q1->x );
	direction.z = 1.0f - 2.0f * ( q1->x * q1->x + q1->y * q1->y );

	vec_normalize(&direction);
	
	return direction;
}

/*! \brief multiply two quaternions, result in q0 */
void quat_mult( quat_t *q0, quat_t *q1 )
{
	GLfloat rx , ry, rz, rw;

	rw = q0->w * q1->w - q0->x * q1->x - q0->y * q1->y - q0->z * q1->z;
	rx = q0->w * q1->x + q0->x * q1->w + q0->y * q1->z - q0->z * q1->y;
	ry = q0->w * q1->y + q0->y * q1->w + q0->z * q1->x - q0->x * q1->z;
	rz = q0->w * q1->z + q0->z * q1->w + q0->x * q1->y - q0->y * q1->x;

	q0->w = rw;
	q0->x = rx;
	q0->y = ry;
	q0->z = rz;
}

/*! \brief return the matrix from a quaternion */
void quat_tomat(quat_t *quat, GLfloat *matrix)
{
	GLfloat xx, yy, zz, xy, wz, xz, wy, yz, wx;	

	xx = quat->x * quat->x;
	yy = quat->y * quat->y;
	zz = quat->z * quat->z;
	xy = quat->x * quat->y;
	wz = quat->w * quat->z;
	xz = quat->x * quat->z;
	wy = quat->w * quat->y;
	yz = quat->y * quat->z;
	wx = quat->w * quat->x;

	matrix[  0 ] = 1.0f - 2.0f * ( yy + zz );
	matrix[  1 ] = 2.0f * ( xy + wz );
	matrix[  2 ] = 2.0f * ( xz - wy );
	matrix[  3 ] = 0.0f;
	matrix[  4 ] = 2.0f * ( xy - wz );
	matrix[  5 ] = 1.0f - 2.0f * ( xx + zz );
	matrix[  6 ] = 2.0f * ( yz + wx );
	matrix[  7 ] = 0.0f;
	matrix[  8 ] = 2.0f * ( xz + wy );
	matrix[  9 ] = 2.0f * ( yz - wx );
	matrix[ 10 ] = 1.0f - 2.0f * ( xx + yy );
	matrix[ 11 ] = 0.0f;
	matrix[ 12 ] = 0.0f;
	matrix[ 13 ] = 0.0f;
	matrix[ 14 ] = 0.0f;
	matrix[ 15 ] = 1.0f;
}

/*! \brief quaternion spherical interpolation
 *
 * spherical interpolation of two quaternion \n
 * help to implement from : \n
 * http://www.euclideanspace.com/maths/algebra/realNormedAlgebra/quaternions/slerp/
 */
quat_t quat_slerp(quat_t *q1, quat_t *q2, float u)
{
	quat_t qslerp;
	GLfloat cos_half_theta, half_theta, sin_half_theta, v1, v2;

	cos_half_theta = q2->w*q1->w + q2->x*q1->x + q2->y*q1->y + q2->z*q1->z ;

	/* acos(x) with abs(x) >= 1.0 is undefined */
	if( abs(cos_half_theta) >= 1 )
	{
		qslerp.x = q1->x;
		qslerp.y = q1->y;
		qslerp.z = q1->z;
		qslerp.w = q1->w;
		return qslerp;
	}

	/* */
	if(cos_half_theta < 0)
		cos_half_theta = -cos_half_theta;

	half_theta = acos(cos_half_theta);

	/* same as sin_half_theta = sin(half_theta) */
	sin_half_theta = sqrt(1.0f - cos_half_theta * cos_half_theta);
	
	/* if theta = 180, result is not fully defined */
	/* rotate around any axis normal to q1 or q2 */
	if( fabs(sin_half_theta) < 0.001f )
	{ 		
		qslerp.w = (q1->w * 0.5 + q2->w * 0.5);
		qslerp.x = (q1->x * 0.5 + q2->x * 0.5);
		qslerp.y = (q1->y * 0.5 + q2->y * 0.5);
		qslerp.z = (q1->z * 0.5 + q2->z * 0.5);
		return qslerp;
	}

	v1 = sin( (1.0-u)*half_theta ) / sin_half_theta;
	v2 = sin(u*half_theta) / sin_half_theta;

	qslerp.x = (v1 * q1->x) + (v2 * q2->x);
	qslerp.y = (v1 * q1->y) + (v2 * q2->y);
	qslerp.z = (v1 * q1->z) + (v2 * q2->z);
	qslerp.w = (v1 * q1->w) + (v2 * q2->w);

	return qslerp;
}

/*! \brief quaternion linear interpolation */
quat_t quat_nlerp(quat_t *q1, quat_t *q2, float u)
{
	quat_t qnlerp;

	qnlerp.x = q1->x + (u * (q2->x - q1->x));
	qnlerp.y = q1->y + (u * (q2->y - q1->y));
	qnlerp.z = q1->z + (u * (q2->z - q1->z));
	qnlerp.w = q1->w + (u * (q2->w - q1->w));

	return qnlerp;
}
