/*
* Copyright dylhoxic
*
* dylhoxic _at_ tuxfamily _dot_ org
*
* This software is a first person shooter game.
*
* This software is governed by the CeCILL version 2 license under French law and
* abiding by the rules of distribution of free software.  You can  use, 
* modify and/ or redistribute the software under the terms of the CeCILL
* license as circulated by CEA, CNRS and INRIA at the following URL
* "http://www.cecill.info". 
*
* As a counterpart to the access to the source code and rights to copy,
* modify and redistribute granted by the license, users are provided only
* with a limited warranty  and the software's author,  the holder of the
* economic rights,  and the successive licensors  have only  limited
* liability. 
*
* In this respect, the user's attention is drawn to the risks associated
* with loading,  using,  modifying and/or developing or reproducing the
* software by the user in light of its specific status of free software,
* that may mean  that it is complicated to manipulate,  and  that  also
* therefore means  that it is reserved for developers  and  experienced
* professionals having in-depth computer knowledge. Users are therefore
* encouraged to load and test the software's suitability as regards their
* requirements in conditions enabling the security of their systems and/or 
* data to be ensured and,  more generally, to use and operate it in the 
* same conditions as regards security. 
*
* The fact that you are presently reading this means that you have had
* knowledge of the CeCILL version 2 license and that you accept its terms.
*/

/*! \file shader.c
* \brief shader functions
*/

#include "shader.h"

/*! \brief populate a glsl program objects with a vertex && fragment shader */
GLboolean shader_create(sh_glsl_t *shader, char *vs_filename, char *fs_filename)
{
	if(!shader_compile(&(shader->program), &(shader->vertex_shader), GL_VERTEX_SHADER_ARB, vs_filename ))
		return GL_FALSE; 
	if(!shader_compile(&(shader->program), &(shader->fragment_shader), GL_FRAGMENT_SHADER_ARB, fs_filename ))
		return GL_FALSE;
	return GL_TRUE;
}

/*! \brief load a vertex or a fragment shader shader 
 *
 * and attach it to the program object
 * also perform program initialization and linking
 */
GLboolean shader_compile(GLhandleARB *program, GLhandleARB *shader, GLenum shader_type, char *filename)
{
	GLcharARB *shader_src;
	GLint infos;
	/* create shader object */
	*shader = _glCreateShaderObjectARB(shader_type);
	
	/* get shader from file */
	shader_src = shader_from_file(filename);

	if(shader_src==NULL)
	{
		fprintf(stderr, "Could not load file %s : %s\n", filename, SDL_GetError());
		return GL_FALSE;
	}
	
	/* attach the source to the shader object */
	_glShaderSourceARB(*shader, 1, &shader_src, NULL);

	free(shader_src);
	shader_src = NULL;

	/* compile it */
	_glCompileShaderARB(*shader);

	/* get compile infos */
	_glGetObjectParameterivARB(*shader, GL_OBJECT_COMPILE_STATUS_ARB, &infos);	
	if( infos == 1)
	{
		#if HKW_DEBUG_SHADER
		fprintf(stdout, "shader '%s' compilation ok\n", filename);
		#endif
	}
	else
	{
		fprintf(stderr, "shader %s compilation error\n", filename);
		#if HKW_DEBUG_SHADER
		shader_get_log(shader);
		#endif
		return GL_FALSE;
	}

	if(shader_type==GL_VERTEX_SHADER_ARB)
		*program = _glCreateProgramObjectARB();

	/* attach the shader to our program */
	_glAttachObjectARB(*program, *shader);

	if(shader_type==GL_FRAGMENT_SHADER_ARB)
	{
		_glLinkProgramARB(*program);

		_glGetObjectParameterivARB(*program, GL_OBJECT_LINK_STATUS_ARB, &infos);	
		/*
		if( infos == 1)
		{	
			#if HKW_DEBUG_SHADER
			fprintf(stdout, "shader program succesfully linked\n");
			#endif
		}
		else
		{
			fprintf(stderr, "shader program link error\n");
			#if HKW_DEBUG_SHADER
			shader_get_log(program);
			#endif
			return GL_FALSE;
		}
		*/
		if( infos != 1 )
		{
			fprintf(stderr, "shader program link error\n");
			#if HKW_DEBUG_SHADER
			shader_get_log(program);
			#endif
			return GL_FALSE;
		}

		#if HKW_DEBUG_SHADER
		fprintf(stdout, "shader program succesfully linked\n");
		#endif

	}
	
	return GL_TRUE;
}

/*! \brief initialize && allocate a shader */
GLboolean shader_init()
{
	#if HKW_DEBUG_SHADER
	GLint infos;
	#endif
	
	/* the four following extensions are required for glsl shader */
	if( !find_ext("GL_ARB_shading_language_100"))
	{
		SDL_SetError("GL_ARB_shading_language_100 not found\n");
		return GL_FALSE;
	}

	if( !find_ext("GL_ARB_shader_objects"))
	{
		SDL_SetError("GL_ARB_shader_objects not found\n");
		return GL_FALSE;
	}

	_glCreateShaderObjectARB = (PFNGLCREATESHADEROBJECTARBPROC)SDL_GL_GetProcAddress("glCreateShaderObjectARB");	
	_glShaderSourceARB = (PFNGLSHADERSOURCEARBPROC)SDL_GL_GetProcAddress("glShaderSourceARB");
	_glCompileShaderARB = (PFNGLCOMPILESHADERARBPROC)SDL_GL_GetProcAddress("glCompileShaderARB");
	_glCreateProgramObjectARB = (PFNGLCREATEPROGRAMOBJECTARBPROC)SDL_GL_GetProcAddress("glCreateProgramObjectARB");
	_glAttachObjectARB = (PFNGLATTACHOBJECTARBPROC)SDL_GL_GetProcAddress("glAttachObjectARB");
	_glLinkProgramARB = (PFNGLLINKPROGRAMARBPROC)SDL_GL_GetProcAddress("glLinkProgramARB");
	_glUseProgramObjectARB = (PFNGLUSEPROGRAMOBJECTARBPROC)SDL_GL_GetProcAddress("glUseProgramObjectARB");
	_glGetObjectParameterivARB = (PFNGLGETOBJECTPARAMETERIVARBPROC)SDL_GL_GetProcAddress("glGetObjectParameterivARB");
	_glGetInfoLogARB = (PFNGLGETINFOLOGARBPROC)SDL_GL_GetProcAddress("glGetInfoLogARB");
	_glDetachObjectARB = (PFNGLDETACHOBJECTARBPROC)SDL_GL_GetProcAddress("glDetachObjectARB");
	_glDeleteObjectARB = (PFNGLDELETEOBJECTARBPROC)SDL_GL_GetProcAddress("glDeleteObjectARB");

	_glGetUniformLocationARB = (PFNGLGETUNIFORMLOCATIONARBPROC)SDL_GL_GetProcAddress("glGetUniformLocationARB");
	_glUniformMatrix4fvARB = (PFNGLUNIFORMMATRIX4FVARBPROC)SDL_GL_GetProcAddress("glUniformMatrix4fvARB");

	if( (!_glCreateShaderObjectARB) || (!_glShaderSourceARB) || (!_glCompileShaderARB) ||
		(!_glCreateProgramObjectARB) || (!_glAttachObjectARB) || (!_glLinkProgramARB) ||  
		(!_glUseProgramObjectARB) || (!_glGetObjectParameterivARB) || (!_glGetInfoLogARB) || 
		(!_glDetachObjectARB) || (!_glDeleteObjectARB) || (!_glGetUniformLocationARB ) || (!_glUniformMatrix4fvARB) )
	{
		SDL_SetError("Couldn't get address of one or more functions of GL_ARB_shader_objects\n");
		return GL_FALSE;
	}
	
	if( !find_ext("GL_ARB_vertex_shader"))
	{
		SDL_SetError("GL_ARB_vertex_shader not found\n");
		return GL_FALSE;
	}

	_glGetAttribLocationARB = (PFNGLGETATTRIBLOCATIONARBPROC)SDL_GL_GetProcAddress("glGetAttribLocationARB");
	_glVertexAttrib4fvARB = (PFNGLVERTEXATTRIB4FVARBPROC)SDL_GL_GetProcAddress("glVertexAttrib4fvARB") ;
	_glVertexAttribPointerARB = (PFNGLVERTEXATTRIBPOINTERARBPROC)SDL_GL_GetProcAddress("glVertexAttribPointerARB");
	_glEnableVertexAttribArrayARB = (PFNGLENABLEVERTEXATTRIBARRAYARBPROC)SDL_GL_GetProcAddress("glEnableVertexAttribArrayARB") ;
	_glDisableVertexAttribArrayARB = (PFNGLDISABLEVERTEXATTRIBARRAYARBPROC)SDL_GL_GetProcAddress("glDisableVertexAttribArrayARB") ;
	
	if( (!_glGetAttribLocationARB) || (!_glVertexAttrib4fvARB) || (!_glVertexAttribPointerARB) || 
		(!_glEnableVertexAttribArrayARB) || (!_glDisableVertexAttribArrayARB) )
	{
		SDL_SetError("Couldn't get address of one or more functions of GL_ARB_vertex_shader\n");
		return GL_FALSE;
	}

	if( !find_ext("GL_ARB_fragment_shader"))
	{
		SDL_SetError("GL_ARB_fragment_shader not found\n");
		return GL_FALSE;
	}	

	fprintf(stdout,"GLSL :\n VERSION : %s\n", (char *)glGetString(GL_SHADING_LANGUAGE_VERSION_ARB));
	#if HKW_DEBUG_SHADER
	glGetIntegerv(GL_MAX_VERTEX_ATTRIBS_ARB, &infos);
	fprintf(stdout," MAX_VERTEX_ATTRIBS_ARB : %d\n", infos);
	glGetIntegerv(GL_MAX_VERTEX_UNIFORM_COMPONENTS_ARB, &infos);
	fprintf(stdout," MAX_VERTEX_UNIFORM_COMPONENTS_ARB : %d\n", infos);
	glGetIntegerv(GL_MAX_VARYING_FLOATS_ARB, &infos);
	fprintf(stdout," MAX_VARYING_FLOATS_ARB : %d\n", infos);
	glGetIntegerv(GL_MAX_VERTEX_ATTRIBS_ARB, &infos);
	fprintf(stdout," MAX_COMBINED_TEXTURE_IMAGE_UNITS_ARB : %d\n", infos);
	glGetIntegerv(GL_MAX_VERTEX_TEXTURE_IMAGE_UNITS_ARB, &infos);
	fprintf(stdout," MAX_VERTEX_TEXTURE_IMAGE_UNITS_ARB : %d\n", infos);
	glGetIntegerv(GL_MAX_TEXTURE_IMAGE_UNITS_ARB, &infos);
	fprintf(stdout," MAX_TEXTURE_IMAGE_UNITS_ARB : %d\n", infos);
	glGetIntegerv(GL_MAX_TEXTURE_COORDS_ARB, &infos);
	fprintf(stdout," MAX_TEXTURE_COORDS_ARB : %d\n", infos);
	glGetIntegerv(GL_MAX_FRAGMENT_UNIFORM_COMPONENTS_ARB, &infos);
	fprintf(stdout," MAX_FRAGMENT_UNIFORM_COMPONENTS_ARB : %d\n", infos);
	#endif

	return GL_TRUE;
}

/*! \brief read a file */ 
char *shader_from_file(char *filename)
{
	FILE *fp;
	int size = 0;
	char *shader = NULL;

	/* fp = fopen(filename, "r+"); */
	fp = file_open(filename, "r+");

	if( fp != NULL )
	{
		fseek(fp, 0, SEEK_END);
		size = ftell(fp);

		if((size <= 0) || (size > GLSL_MAX_SIZE))
		{
			SDL_SetError("wrong shader size or size > %d", GLSL_MAX_SIZE);
			return NULL;
		}
		
		fseek(fp, 0, SEEK_SET);
		shader = (char *)malloc(sizeof(char)*(size+1));

		if(shader==NULL)
		{
			SDL_SetError("Could not alloc data for shader\n");
			return NULL;
		}

		size = fread(shader,sizeof(char),size,fp);
		shader[size] = '\0';

		fclose(fp);
		
		#if HKW_DEBUG_SHADER
		/* fprintf(stdout, "shader size : %d\n shader src :\n %s \n", size, shader); */
		#endif
		
		return shader;

	}
	else
	{
		SDL_SetError("Could not open file %s", filename);
		return NULL;
	}
}

/*! \brief get shader log and print it on stderr */
void shader_get_log(GLhandleARB *obj)
{
	GLint shader_length = 0;
	GLint shader_written_chars = 0;
	char *shader_log = NULL;

	_glGetObjectParameterivARB(*obj, GL_OBJECT_INFO_LOG_LENGTH_ARB, &shader_length);
	
	if( shader_length > 0) 
	{
		shader_log = (char *)malloc(sizeof(char)*shader_length);
		if(shader_log != NULL) 
		{
			_glGetInfoLogARB(*obj, shader_length, &shader_written_chars, shader_log);
			if(shader_written_chars > 0) 
			{
				fprintf(stderr, "written chars : %d\n", shader_written_chars);
				fprintf(stderr, "log :\n");
				fprintf(stderr, "%s\n", shader_log);
			}
			free(shader_log);
		}
	}
}

/*! \brief free a shader */
void shader_free(sh_glsl_t *shader)
{
	_glDetachObjectARB(shader->program, shader->vertex_shader);
	_glDetachObjectARB(shader->program, shader->fragment_shader);
	_glDeleteObjectARB(shader->vertex_shader);
	_glDeleteObjectARB(shader->fragment_shader);
	_glDeleteObjectARB(shader->program);
}


